import numpy as np 
def comp_sh_vars(lon: float, r: float, theta: float, sh_degrees: int) -> dict[str, list[float]]:
    """
    Computes Spherical Variabale and put the results in a dictionary
    variable computed are (a/r) ^ (n+2), cos_m(lon), and sin_m(lon)

    Parameters:
    lon (float or int): longitude in degrees
    r (float or int): distance from the center of ellipsoid computed by geod_to_geoc function
    theta (float or int): geocentric latitude computed by geod_to_geoc function
    sh_degrees(int): Total number of spherical harmonic degrees

    Returns:
    dictionary: with keys relative_radius_power, cos_mlon, and sin_mlon for the three different value arrays
    """
    ellip_re = 6371.2 # Mean radius of ellipsoid

    sh_degrees = int(sh_degrees)
    sh_vars = {}
    cos_lon = np.cos(np.radians(lon))
    sin_lon = np.sin(np.radians(lon))

    sh_vars["relative_radius_power"] = []

    ellip_re_over_r = ellip_re/r

    sh_vars["relative_radius_power"].append((ellip_re_over_r) ** 2)

    for n in range(1,sh_degrees+1):
        sh_vars["relative_radius_power"].append(sh_vars["relative_radius_power"][n-1] * ellip_re_over_r)

    sh_vars["cos_mlon"] = []
    sh_vars["sin_mlon"] = []

    sh_vars["cos_mlon"].append(1.0)
    sh_vars["sin_mlon"].append(0.0)

    sh_vars["cos_mlon"].append(cos_lon)
    sh_vars["sin_mlon"].append(sin_lon)

    for m in range(2, sh_degrees+1):
        sh_vars["cos_mlon"].append(sh_vars["cos_mlon"][m-1] * cos_lon -  sh_vars["sin_mlon"][m-1] * sin_lon)
        sh_vars["sin_mlon"].append(sh_vars["cos_mlon"][m-1] * sin_lon +  sh_vars["sin_mlon"][m-1] * cos_lon)

    return sh_vars





import math
from typing import Union

import numpy as np
import warnings

def Flattened_Chaos_Legendre1(nmax: int, theta: Union[float, list[float]], epsilon: float = 1e-6, silence_warnings: bool = False) -> list[list[float]]:
    """
    Inputs:
    nmax (int): maximum degree of legendre polynomial (all orders for each degree are calculated)
    theta (float or np.array(float)): colatitude values where the legendre polynomials are calculated
    epsilon (float): By how much pole values are shifted away from poles to avoid divide by zero (Should be > 1e-6)
    silence_warnings (bool): True -> do not display warning that theta values < epsilon have been shifted. False -> display

    Outputs:
    Takes nmax and colatitude (degrees) as inputs
    Outputs a 2 dimensional numpy array which contains the associated
    legendre polynomials (Pnm) and the respective derivatives (dPnm).
    In these respective arrays the values are arranged in order major order:
    i.e.
    P10,P20,P30...Pnmax0, P11,P21,P31...Pnmax1
    """

    if np.isscalar(theta):
        theta = np.array([theta])
    if (np.isclose(0, min(theta), epsilon)) or (np.isclose(max(theta), 180, epsilon)):
        if not silence_warnings:
            warnings.warn(f'Input coordinates include the poles. They have been shifted by {epsilon}')
        mask_0 = np.isclose(theta, 0, atol=epsilon)
        mask_180 = np.isclose(theta, 180, atol=epsilon)
        theta[mask_0 | mask_180] = epsilon

    if min(theta) < 0.0 or max(theta) > 180.0:
        raise ValueError('Colatitude outside bounds [0, 180].')

    costh = np.cos(np.radians(theta))
    sinth = np.sqrt(1 - costh * costh)

    Pnm = []
    dPnm = []
    Pnm.append(np.ones(len(costh)))  # is copied into trailing dimensions
    dPnm.append(np.zeros(len(costh)))

    rootn = np.sqrt(np.arange(2 * nmax ** 2 + 1))

    # Recursion relations after Langel "The Main Field" (1987),
    # eq. (27) and Table 2 (p. 256)
    for m in range(nmax):
        if m == 1:
            Pnm.append(sinth)
            dPnm.append(costh)
        # Buffer normalization factor for [n,n] to [n,n-1]
        c2 = rootn[m + m + 1]
        # Pnm_tmp holds Pnm[m,m] with normalization factor
        Pnm_tmp = c2 * Pnm[-1]
        # Pnm[m,m-1] = Pnm[m-1,m-1] * cos * c2
        Pnm.append(costh * Pnm_tmp)
        # Hold this for the derivative of diagonal later
        dPnm_diag_tmp = Pnm[-1]
        # Derivative of previous diagonal * cos - sin * previous diagonal
        dPnm.append(dPnm[-1] * costh * c2 - sinth * Pnm_tmp)

        for n in range(m + 2, nmax + 1):
            d = n * n - m * m
            e = n + n - 1

            Pnm.append((e * costh * Pnm[-1] - rootn[d - e] * Pnm[-2]) / rootn[d])
            dPnm.append((n * costh * Pnm[-1] - rootn[d] * Pnm[-2]) / sinth)

        if m > 0:  # Diagonal append Pnm[m,m] = sin^m(theta)
            Pnm.append(sinth * Pnm_tmp / rootn[m + m + 2])
            dPnm.append((dPnm_diag_tmp * rootn[m + 1] * np.sqrt(0.5)))
    if nmax == 1:
        Pnm.append(sinth)
        dPnm.append(costh)
    return [Pnm, dPnm]

def get_index(n: int, m:int, nmax: int) -> int:
    """
    Get the index of Flattened_Chaos_Legendre1
    Args:
        n: degree number
        m: order number
        nmax: maximum degree or order

    Returns:
        The index of legendre polynomial value in Flattened_Chaos_Legendre1

    The order of legendre polynomial in Flattened_Chaos_Legendre1()
        {degree}{order}
        00 (degree=0, order=0)
        10
        20
        30
        ...
        {nmax}0
        11
        21
        ...
        {nmax}1
        22
        32
        ...

    """
    if m == 0:
        return n
    else:
        return int(m * ((nmax + 1) + (nmax + 1 - (m - 1))) // 2 + (n - m))




import math
import numpy as np
import copy

def calc_sh_degrees_to_num_elems(sh_degrees):
    """
    Gives you the number of elements from the number of 
    sphereical harmonic degree

    Parameters:
    sh_degrees (int): The number of spherical harmonic degrees

    Returns:
    int: The number of elements
    """
    return int(sh_degrees * (sh_degrees+3)/2 ) + 1

def calc_num_elems_to_sh_degrees(num_elems):
    """
    Gives you the number of spherical harmonic degrees from the total number
    of elements

    Parameters:
    num_elems (int): The number of elements

    Returns:
    int: The number of spherical harmonic degrees
    """
    return int((-3+np.sqrt(8*num_elems + 1) )/2)


def load_coef(filename, skip_two_columns=False, load_sv=True, end_degree=None, load_year=None):
    """
    Takes a coefficient filename or path and gives you back a dictionary with the
    components in arrays under the keys g,h,g_sv, and h_sv

    Parameters:
    filename (string): The relative path or just the name of the coefficient file
    skip_two_columns (boolean): Sometimes the coefficient file in the data lines
    lists the spherical harmonic degree numbers in the first two columns like in
    WMM and IGRF. If this parameter is true it lets the function know to skip those
    two columns
    load_sv (boolean): The coefficient file doesn't always have sv columns like
    in HDGM crust,set this false to let the function know to not load g_sv and h_sv
    end_degree (int): If you just want to load a specific amount of spherical harmonic
    degrees from the coefficient file set this to that number. Set it to None if
    you wish to ignore
    load_year (int): The coefficient file could be seperated by year components like
    in HDGM core, set this to a year value based on what year you want to be loaded.
    Set it to None if you wish to ignore this parameter

    Returns:
    dictionary: The dictionary loaded with g, h, g_sv, h_sv arrays under those keys
    """
    coef_dict = {"g": [], "h": []}
    if load_sv:
        coef_dict["g_sv"] = []
        coef_dict["h_sv"] = []

    skip_adder = 0
    if skip_two_columns:
        skip_adder = 2

    coef_file = open(filename, 'r')
    lines = coef_file.readlines()

    header_line = True
    load = True
    footer_line_split_len = 4
    if not load_sv:
        footer_line_split_len = 2
    if load_year is not None:
        load = False
    num_lines_load = None
    load_counter = 0
    if end_degree is not None:
        num_lines_load = calc_sh_degrees_to_num_elems(end_degree)
    for line in lines:
        split = line.split()
        # This will detect the footer and we can stop loading
        if len(split) < footer_line_split_len:
            break

        if load_year is not None and (split[1] == (str(load_year) + ".0") or split[1] == (str(load_year) + ".00")):
            load = True
            header_line = False
            coef_dict["epoch"] = int(float(split[1]))
            continue
        if load_year is not None and (split[1] == (str(load_year) + ".0") or split[1] == (str(load_year+5) + ".00")):
            break
        if header_line:
            header_line = False
            if load_sv:
                coef_dict["epoch"] = int(float(split[1]))
            continue
        if num_lines_load is not None and load_counter >= num_lines_load:
            break
        if load:
            load_counter = load_counter + 1
            coef_dict['g'].append(float(split[0 + skip_adder]))
            coef_dict['h'].append(float(split[1 + skip_adder]))
            if load_sv:
                coef_dict['g_sv'].append(float(split[2 + skip_adder]))
                coef_dict['h_sv'].append(float(split[3 + skip_adder]))

    coef_file.close()

    if len(coef_dict["g"]) > 0 and (coef_dict["g"][0] != 0 or coef_dict['h'][0] != 0):
        coef_dict["g"].insert(0, 0)
        coef_dict["h"].insert(0, 0)
        if load_sv:
            coef_dict["g_sv"].insert(0, 0)
            coef_dict["h_sv"].insert(0, 0)

    if end_degree is not None and len(coef_dict["g"]) > num_lines_load:
        coef_dict["g"].pop()
        coef_dict["h"].pop()
        if load_sv:
            coef_dict["g_sv"].pop()
            coef_dict["h_sv"].pop()

    return coef_dict
def timely_modify_magnetic_model(sh_dict, dec_year):
    """
    Time change the Model coefficients from the base year of the model(epoch) using secular variation coefficients.
Store the coefficients of the static model with their values advanced from epoch t0 to epoch t.
Copy the SV coefficients.  If input "t�" is the same as "t0", then this is merely a copy operation.

    Parameters:
    sh_dict (dictionary): This is the input dictionary, you would get this dictionary from using the load_coef function
    dec_year(float or int): Decimal year input for calculating the time shift
    epoch (float or int): The base year of the model

`   Returns:
    dictionary: Copy of sh_dict with the elements timely shifted
    """

    sh_dict_time = copy.deepcopy(sh_dict)
    epoch = sh_dict.get("epoch", 0)
    #If the sh_dict doesn't have secular variations just return a copy
    #of the dictionary
    if  "g_sv" not in sh_dict or "h_sv" not in sh_dict:
        return sh_dict_time
    num_elems = len(sh_dict["g"])
    sh_degrees = calc_num_elems_to_sh_degrees(num_elems)
    date_diff = dec_year - epoch
    for n in range(1, (sh_degrees+1)):
        for m in range(n+1):
            index = int(n * (n + 1) / 2 + m)
            if index < num_elems:
                sh_dict_time["g"][index] = sh_dict["g"][index] + date_diff * sh_dict["g_sv"][index]
                sh_dict_time["h"][index] = sh_dict["h"][index] + date_diff * sh_dict["h_sv"][index]

    return sh_dict_time






import sys
import os
import unittest
import numpy as np
import math

# Add the parent directory to the system path
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from geomaglib import util

class TestUtil(unittest.TestCase):

    def _test_single_geod_to_geoc_conv(self, lat, alt, exp_r, exp_theta):
        act_r, act_theta = util.geod_to_geoc_lat(lat,alt)
        self.assertAlmostEqual(act_r, exp_r, places=10)
        self.assertAlmostEqual(act_theta, exp_theta,places=10)

    def test_geod_to_geoc_conv(self):
        self._test_single_geod_to_geoc_conv(0,0,6378.137,0)
        self._test_single_geod_to_geoc_conv(90,0,6356.75231424518,90)
        self._test_single_geod_to_geoc_conv(-90,2,6358.75231424518,-90)
        self._test_single_geod_to_geoc_conv(22.342,2.02, 6377.08899029245,22.2070535610)
        self._test_single_geod_to_geoc_conv(45.734,-5,6362.21562444842,45.5414722008)
        self._test_single_geod_to_geoc_conv(-36.254,3.45,6374.14924310438,-36.0707588822) 
        self._test_single_geod_to_geoc_conv(-70.89,-7,6352.06156995608,-70.7705043304) 
        self._test_single_geod_to_geoc_conv(50.5,0,6365.44775048439,50.3109909949)
        self._test_single_geod_to_geoc_conv(10.2,20,6397.47181557818,10.1333459251)
    
    def test_alt_to_ellipsoid(self):

        self.assertAlmostEqual(util.alt_to_ellipsoid_height(10,40,20),10.034187,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(5,20,60),4.957280,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(1,-90,0),0.970466,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(0,90,0),0.013606,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(-2,45,-50),-1.982954,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(5,-60,-50),5.019614,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(3.5,20,270),3.489277,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(2.005,5.5,40)[0],1.985737,places=10)
        self.assertAlmostEqual(util.alt_to_ellipsoid_height(7,-20,-20),6.999373,places=10)
    
    def test_calc_dec_year(self):
        self.assertAlmostEqual(util.calc_dec_year(2024,5,20),2024.3825136612,places=10)
        self.assertAlmostEqual(util.calc_dec_year(2023,5,20),2023.3808219178,places=10)
        self.assertAlmostEqual(util.calc_dec_year(2022,2,11),2022.11232876712,places=10)
        self.assertAlmostEqual(util.calc_dec_year(2021,8,30),2021.6602739726,places=10)
        self.assertAlmostEqual(util.calc_dec_year(2020,2,11),2020.11202185792,places=10)
        self.assertAlmostEqual(util.calc_dec_year(2015,10,1),2015.74794520547,places=10)
        self.assertAlmostEqual(util.calc_dec_year(2011,12,31),2011.99726027397,places=10)
        self.assertAlmostEqual(util.calc_dec_year(1900,3,13),1900.19452054794,places=10)

    def test_calc_dec_year_arr(self):

        years = np.array([2024, 2023, 2022, 2021, 2020, 2015, 2011, 1900])
        months = np.array([5, 5, 2, 8, 2, 10, 12, 3])
        days = np.array([20, 20, 11, 30, 11, 1, 31, 13])

        dyears = util.calc_dec_year_array(years, months, days)


        for i in range(len(years)):
            self.assertAlmostEqual(dyears[i], util.calc_dec_year(years[i], months[i], days[i]), places=10)


    def test_jd2000(self):

        year = 2024
        month = 12
        day = 31
        ut = 12
        minutes = 30
        jd = util.jd2000(year, month, day, ut, minutes)

        self.assertAlmostEqual(jd, 9131.52083333, places=6)











if __name__ == '__main__':
    unittest.main()



from typing import Union

# Autogenerated with SMOP 0.29
#Modified by Adam
import numpy as np
import math
from geomaglib import geoid
import calendar
import datetime as dt


def geod_to_geoc_lat(lat: float, alt: float) -> tuple[float, float]:
    # [r, theta] = geod2geoc(alpha, h);
    # [r, theta, B_r, B_theta] = geod2geoc(alpha, h, X, Z);
    # conversion from geodetic X,Z components to geocentric B_r, B_theta
    # Input:   geodetic latitude lat (deg)
    #          altitude alt [km]
    # Output:  theta (deg)
    #          r (km)

    # Nils Olsen, DSRI Copenhagen, September 2001.
    # After Langel (1987), eq. (52), (53), (56), (57)
    # Converted to python by Adam (2017)

    # Latest changes: NIO, 170903 optional conversion of magnetic components
    #                 Update to WGS-84 ellipsoid

    # WGS-84 Ellipsoid parameters
    a = 6378.137
    f = 1 / 298.257223563
    b = a * (1 - f)
    lat_rad = np.radians(lat)
    sin_alpha_2 = np.sin(lat_rad) ** 2
    cos_alpha_2 = np.cos(lat_rad) ** 2
    tmp = np.multiply(alt, np.sqrt(np.dot(a ** 2, cos_alpha_2) + np.dot(b ** 2, sin_alpha_2)))
    beta = np.arctan(np.multiply((tmp + b ** 2) / (tmp + a ** 2), np.tan(lat_rad)))
    theta = np.rad2deg(beta)
    r = np.sqrt(alt ** 2 + np.dot(2, tmp) + np.dot(a ** 2, (1 - np.dot((1 - (b / a) ** 4), sin_alpha_2))) / (
                1 - np.dot((1 - (b / a) ** 2), sin_alpha_2)))

    return r, theta


def alt_to_ellipsoid_height(alt: Union[float, np.float64, list, np.ndarray], lat: Union[float, np.float64, list, np.ndarray], lon: Union[float, np.float64, list, np.ndarray]) -> np.ndarray:
    """
    This function converts altitude MSL in kilometers to ellipsoid height in
    kilometers

    Parameters:
    alt (int or float): The original altitude in kilometers in MSL
    lat (int or float): The latitude degree from -90 to 90
    lon (int or float): The longitude degree from -180 to 360

    Returns:
    (float): Height above ellipsoid in kilometers
    """
    if np.isscalar(lat):
        lat = np.array([lat])
    if np.isscalar(lon):
        lon = np.array([lon])
    if np.isscalar(alt):
        alt = np.array([alt])

    if isinstance(lat, list):
        lat = np.array(lat)
    if isinstance(lon, list):
        lon = np.array(lon)
    if isinstance(alt, list):
        alt = np.array(alt)
    offset_x = np.zeros_like(lon, dtype=float)
    offset_y = np.zeros_like(lon, dtype=float)
    lon_mask = lon < 0
    offset_x[lon_mask] = (lon[lon_mask] + 360) * geoid.geoid["scale_factor"]
    offset_x[~lon_mask] = lon[~lon_mask] * geoid.geoid["scale_factor"]   
    # if lon < 0.0:
    #     offset_x = (lon + 360) * geoid.geoid["scale_factor"]
    # else:
    #     offset_x = lon * geoid.geoid["scale_factor"]

    offset_y = (90 - lat) * geoid.geoid["scale_factor"]

    # post_x = int(np.floor(offset_x))
    # if post_x + 1 == geoid.geoid["cols"]:
    #     post_x = post_x - 1

    # post_y = int(np.floor(offset_y))
    # if post_y + 1 == geoid.geoid["rows"]:
    #     post_y = post_y - 1

    post_x = np.floor(offset_x).astype(int)
    post_x = np.where(post_x + 1 == geoid.geoid["cols"], post_x - 1, post_x)

    post_y = np.floor(offset_y).astype(int)
    post_y = np.where(post_y + 1 == geoid.geoid["rows"], post_y - 1, post_y)

    index = np.asarray(post_y * geoid.geoid["cols"] + post_x)

    elevation_NW = np.array([geoid.geoid["height_buffer"][i] for i in index])
    elevation_NE = np.array([geoid.geoid["height_buffer"][i + 1] for i in index])
    index = (post_y + 1) * geoid.geoid["cols"] + post_x
    elevation_SW = np.array([geoid.geoid["height_buffer"][i] for i in index])
    elevation_SE = np.array([geoid.geoid["height_buffer"][i + 1] for i in index])

    
    # elevation_SW = geoid.geoid["height_buffer"][index]
    # elevation_SE = geoid.geoid["height_buffer"][index + 1]

    delta_x = offset_x - post_x
    delta_y = offset_y - post_y

    upper_y = elevation_NW + delta_x * (elevation_NE - elevation_NW)
    lower_y = elevation_SW + delta_x * (elevation_SE - elevation_SW)

    delta_height = upper_y + delta_y * (lower_y - upper_y)

    # delta_height is in meters need to convert to kilometers
    ellipsoid_height = alt + delta_height / 1000

    return ellipsoid_height

def calc_dec_year(year: int, month: int, day:int ) -> float:
    """
    Takes year, month, and day and calculates the decimal year from those inputs

    Parameters:
    year (int): The year fully written for example 2024
    month (int): The month from 1-12 where is 1 is January and 12 is December
    day (int): The day of the month from 1-31

    Returns:
    (float): The decimal year

    """
    num_days_year = 365
    if calendar.isleap(year):
        num_days_year = 366
    date = dt.datetime(year, month, day)
    day_of_year = date.timetuple().tm_yday
    return year + ((day_of_year - 1) / num_days_year)

def calc_dec_year_array(year: np.ndarray, month: np.ndarray, day:np.ndarray ) -> np.ndarray:
    """
    Takes year, month, and day and calculates the decimal year from those inputs

    Parameters:
    year (int): The year fully written for example 2024
    month (int): The month from 1-12 where is 1 is January and 12 is December
    day (int): The day of the month from 1-31

    Returns:
    (float): The decimal year
    adjusted calc_dec_year to accept vectors by Collin 01/25 
    """
    dec_year = []
    # if(year.size == 1):
    #     num_days_year = 365
    #     if calendar.isleap(year):
    #         num_days_year = 366
        
    #     date = dt.datetime(year, month, day)
    #     day_of_year = date.timetuple().tm_yday
    #     dec_year.append(year + ((day_of_year - 1) / num_days_year))
    # else:
        
    for i in range(0,year.size):
        num_days_year = 365
        if calendar.isleap(year[i]):
            num_days_year = 366
        
        date = dt.datetime(year[i], month[i], day[i])
        day_of_year = date.timetuple().tm_yday
        dec_year.append(year[i] + ((day_of_year - 1) / num_days_year))

    return np.array(dec_year)
def decimalYearToDateTime(dyear:float) ->tuple[float, float, float, float, float]:

    beginYear = dt.datetime(int(dyear),1,1)
    yearFrac = dyear - int(dyear)
    numDays = (dt.date(int(dyear)+1,1,1)-dt.date(int(dyear),1,1)).days
    theDate = dt.timedelta(days=yearFrac*numDays) + beginYear
    return theDate.year, theDate.month, theDate.day, theDate.hour, theDate.minute

def jd2000(year: int, month: int, day:int , ut: int, minutes: int) -> float:
    """
    Compute the julian day with offset 2000
    Args:
        year: yyyy
        month: 1 -12 month
        day: The day of the month
        ut: The hour from 1- 24
        minutes: The minutes in an hour 0 - 60

    Returns:
        The number of day from 0000 offset by the day since 2000
    """

    offset = dt.datetime(2000,1,1)

    diff = dt.datetime(year, month, day, ut, minutes) - offset

    total_days = diff.days + diff.seconds / (3600.0 * 24)

    return total_days



import numpy as np
import math
from collections import defaultdict
import warnings


#Leg_SHA_optimized_ind
def Leg_SHA_optimized_ind(rho, theta, phi, nmax, mmax, get='all', epsilon = 1e-2, silence_warnings = False):
    """
    Written by Collin Kadlecek 7/2024 Collin.kadlecek@noaa.gov
    Computes matrices to connect the radial, colatitude and azimuthal field
    components to the magnetic potential field in terms of spherical harmonic
    coefficients (Schmidt quasi-normalized). In degree order 
    i.e. g10,g11, h11, g20, g21, h21, g22, h22, ... ,g_nmax_nmax h_nmax_nmax

    Parameters
    ----------

    radius : ndarray, shape (...)
        Array containing the radius in earth radii.
    theta : ndarray, shape (...)
        Array containing the colatitude in degrees
        :math:`[0^\\circ,180^\\circ]`.
    phi : ndarray, shape (...)
        Array containing the longitude in degrees.
    nmax : int, positive
        Maximum degree of the sphercial harmonic expansion.
    mmax : int, positive
        Maximum order of the spherical harmonic expansion (defaults to
        ``nmax``). For ``mmax = 0``, for example, only the zonal terms are
        returned.
    get : {'r', 'all'}, optional
        Return just the radial component of the field, or all components

    Returns
    -------
    A_radius, A_theta, A_phi : ndarray, shape (3, nmax*(nmax+2), len(theta))
        Matrices for radial, colatitude and azimuthal field components. The
        second dimension ``M`` varies depending on ``nmax``, ``nmin`` and
        ``mmax``.

    Algorithm description
    --------
    The implimentation of this function follows the structure of the associated
    legendre polynomial generation function (legendre_poly) created by Nils Olsen from the 
    Chaos magpy package. This function recursively generates the associated legendre
    polynomials. Leg_SHA follows this recursive generations structure. However, it 
    also generates each polynomials derivative for each n,m index. Further, only
    the necessary legendre polynomials are held in memory. Upon each polynomials
    generation, it's associated matrix coefficients are calculated. 

    There are two recursion relations used for the polynomial generation. These
    are equations 26 and 27 from The Main GeoMagnetic Field Langel Chapter 4. 26 is 
    used to generate diagonal (n == m) indeces, and 27 for all other. 

    A different source is used to generate the derivatives. There are two recursion 
    relations which follow the structures of the polynomial generation:
    NOTE:
    dP[l,k] refers to the derivative of legendre polynomial with degree = l and order = k
    (where degree >= order)
    costh = cos(theta) sinth = sin(theta)

    For off diagonal terms (m = n-1) which is a special case of 27:
    dP[n,n-1] = dP[n-1,n-1]*costh*sqrt(n+n+1)-sinth*P[n-1,n-1]*sqrt(n+n+1)
    
    For diagonal terms (m = n) following 26:
    dP[n,n] = P[n,n-1]*sqrt(n)/sqrt(2)

    For all other terms:
    dP[n,m] = (n*costh*P[n,m] - sqrt(n**2-m**2)*P[n-1,m])/sinth
    

    Warnings
    --------
    The function can also return the design matrix for the field components at
    the geographic poles, i.e. where ``theta == 0.`` or ``theta == 180.``.
    However, users should be careful when doing this because the vector basis
    for spherical geocentric coordinates,
    :math:`{{\\mathbf{e}_r, \\mathbf{e}_\\theta, \\mathbf{e}_\\phi}}`,
    depends on longitude, which is not well defined at the poles. That is,
    at the poles, any value for the longitude maps to the same location in
    euclidean coordinates but gives a different vector basis in spherical
    geocentric coordinates. Nonetheless, by choosing a specific value for the
    longitude at the poles, users can define the vector basis, which then
    establishes the meaning of the spherical geocentric components. The vector
    basis for the horizontal components is defined as

    .. math::

        \\mathbf{e}_\\theta &= \\cos\\theta\\cos\\phi\\mathbf{e}_x -
            \\cos\\theta\\sin\\phi\\mathbf{e}_y - \\sin\\theta\\mathbf{e}_z\\\\
        \\mathbf{e}_\\phi &= -\\sin\\phi\\mathbf{e}_x +
            \\cos\\phi\\mathbf{e}_y

    Hence, at the geographic north pole as given by ``theta = 0.`` and
    ``phi = 0.`` (chosen by the user), the returned design matrix ``A_theta``
    will be for components along the direction
    :math:`\\mathbf{e}_\\theta = \\mathbf{e}_x` and ``A_phi`` for components
    along :math:`\\mathbf{e}_\\phi = \\mathbf{e}_y`. However,
    if ``phi = 180.`` is chosen, ``A_theta`` will be for components along
    :math:`\\mathbf{e}_\\theta = -\\mathbf{e}_x` and ``A_phi``
    along :math:`\\mathbf{e}_\\phi = -\\mathbf{e}_y`.

    Examples
    --------
    Create the design matrices given 4 locations on the Earth's surface:

    >>> r = 1 = 6371.2km
    >>> theta = np.array([90., 100., 110., 120.])
    >>> phi = 0.
    >>> A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=1)
    >>> A_radius
    array([[ 1.22464680e-16, 2.00000000e+00, 0.00000000e+00],
           [-3.47296355e-01, 1.96961551e+00, 0.00000000e+00],
           [-6.84040287e-01, 1.87938524e+00, 0.00000000e+00],
           [-1.00000000e+00, 1.73205081e+00, 0.00000000e+00]])

    Say, only the axial dipole coefficient is non-zero, what is the magnetic
    field in terms of spherical geocentric components?

    >>> m = np.array([-30000, 0., 0.])  # model coefficients in nT
    >>> Br = A_radius @ m; Br
    array([-3.67394040e-12, 1.04188907e+04, 2.05212086e+04, 3.00000000e+04])
    >>> Bt = A_theta @ m; Bt
    array([-30000. , -29544.23259037, -28190.77862358, -25980.76211353])
    >>> Bp = A_phi @ m; Bp
    array([0., 0., 0., 0.])

    A more complete example is given below:

    .. code-block:: python

        import numpy as np
        from chaosmagpy.model_utils import design_gauss, synth_values

        nmax = 5  # desired truncation degree, i.e. 35 model coefficients
        coeffs = np.arange(35)  # example model coefficients

        # example locations
        N = 10
        radius = 1 = 6371.2 km # Earth's surface
        phi = np.linspace(-180., 180., num=N)  # azimuth in degrees
        theta = np.linspace(1., 179., num=N)  # colatitude in degrees

        # compute design matrices to compute predictions of the
        # field components from the model coefficients, each is of shape N x 35
        A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=nmax)

        # magnetic components computed from the model
        Br_pred = A_radius @ coeffs
        Bt_pred = A_theta @ coeffs
        Bp_pred = A_phi @ coeffs

        # compute the magnetic components directly from the coefficients
        Br, Bt, Bp = synth_values(coeffs, radius, theta, phi)
        np.linalg.norm(Br - Br_pred)  # approx 0.
        np.linalg.norm(Bt - Bt_pred)  # approx 0.
        np.linalg.norm(Bp - Bp_pred)  # approx 0.

    """
    del_indexes = []
    def calc_sha(k,n,m,rn1,Pnm,dPnm, cos_phi_mem,sin_phi_mem, sin_theta):
        k = k - 1
        dPnm = -dPnm
        k = 2*k
        nr = (n+1) * rn1
        tr = rn1*inverse_sinth
        A_r[k] = nr*Pnm*cos_phi_mem
        A_theta[k] = rn1*dPnm*cos_phi_mem
        A_phi[k] = -tr*Pnm*(-m*sin_phi_mem)
        if(m!=0):
            k+=1
            #h terms as in h11, h21, h22 etc...
            A_r[k] = nr*Pnm*sin_phi_mem
            A_theta[k] = rn1*dPnm*sin_phi_mem
            A_phi[k] = -tr*Pnm*m*cos_phi_mem
        else:
            del_indexes.append(k+1)
            A_r[k +1] = np.nan + np.zeros(len(costh))
            A_theta[k +1] = np.nan + np.zeros(len(costh))
            A_phi[k +1] = np.nan + np.zeros(len(costh))

    if get=='r':
        calcTP=False
    else:
        calcTP=True
    N_data=np.size(theta,0)

    theta_min = np.amin(theta)
    theta_max = np.amax(theta)

    # check if poles are included
    if (np.isclose(0,theta_min,epsilon)) or (np.isclose(theta_max,180,epsilon)):
        if(not silence_warnings):
            warnings.warn(f'Input coordinates include the poles. They have been shifted by {epsilon}')
        mask_0 = np.isclose(theta, 0, atol=epsilon)
        mask_180 = np.isclose(theta, 180, atol=epsilon)
        theta[mask_0 | mask_180] = epsilon
        poles = True
    if (theta_min <= 0.0) or (theta_max >= 180.0):
        raise ValueError('Colatitude outside bounds [0, 180].')
    if (np.isclose(theta_max, 3.14, atol= .01)):
        warnings.warn('Please ensure that your inputs are in earth radii (6371.2km) and degrees for both longitude and colatitude')

    if np.isscalar(theta) is not True and theta.ndim > 1:
        dim2=np.size(theta,1)
    else:
        dim2 = 1
    if dim2 > 1:
        theta=theta.transpose()
        phi=phi.transpose()

    costh = np.cos(np.deg2rad(theta))
    sinth = np.sqrt(1-costh*costh)


    inverse_sinth = 1/sinth

    if np.isscalar(rho) is True:
        rho=rho*np.ones(N_data)
    N_nm=mmax*(mmax + 2) + mmax
    #This looks strange, but np.zeros takes a tuple as input, thus the extra paren is necessary
    A_r=np.zeros((N_nm,N_data))
    if(calcTP):
        A_theta=np.zeros((N_nm,N_data))
        A_phi=np.zeros((N_nm,N_data))

    rootn = np.sqrt(np.arange(2 * nmax**2 + 1))
    # Compute the power array
    powers = -(np.arange(nmax) + 3)

    # Raise each element in rho to each power
    r2n = np.transpose(rho[:, np.newaxis] ** powers)
    # Recursion relations after Langel "The Main Field" (1987),
    # eq. (27) and Table 2 (p. 256)
    phi_rad = np.deg2rad(phi)
    Pnm = []
    dPnm = []


    Pnm.append(np.ones(len(costh)))  # is copied into trailing dimensions
    dPnm.append(np.zeros(len(costh)))
    Pnm.append(np.ones(len(costh)))  # is copied into trailing dimensions
    dPnm.append(np.zeros(len(costh)))
    k=0 #Cycles through all the coefficients for degree n, order m model.


    for m in range(nmax):

        cos_phi = np.cos(phi_rad*m)
        sin_phi = np.sin(phi_rad*m)
        if(m == 1):
            Pnm.append(sinth)
            dPnm.append(costh)
            k = 2
            calc_sha(k,1,1,r2n[0],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)


        # Buffer normalization factor for [n-1,n-1] to [n,n-1]
        c2 = rootn[m + m + 1]
        #Pnm_tmp holds Pnm[m,m] with normalization factor
        Pnm_tmp = c2 * Pnm[-1]
        Pnm.append(costh * Pnm_tmp)

        #Hold this for the derivative of diagonal later
        dPnm_diag_tmp = Pnm[-1]
        #Derivative of previous diagonal * cos - sin * previous diagonal
        dPnm.append(dPnm[-1]*costh*c2 - sinth * Pnm_tmp)
        kstart = int((m+1)*(m+2)/2+m)
        calc_sha(kstart,m+1,m,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)

        Pnm.pop(0)
        dPnm.pop(0)
        k = kstart
        for n in range(m+2, nmax+1):
            d = n * n - m * m
            e = n + n - 1
            k = k + n

            Pnm.append((e * costh *  Pnm[-1] - rootn[d-e] * Pnm[-2])
                         / rootn[d])
            #Must be after Pnm is set to be setting derivative with same n
            #dPnm[n,m] = (n*cos*Pnm[n,m] - d^1/2 * Pnm[n-1,m])/sin**2

            dPnm.append((n*costh*Pnm[-1] - rootn[d]* Pnm[-2])*inverse_sinth)
            calc_sha(k,n,m,r2n[n-1],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
            Pnm.pop(0)
            dPnm.pop(0)

        if m > 0:#Diagonal append Pnm[m,m] = sin^m(theta)
            cos_phi = np.cos(phi_rad*(m+1))
            sin_phi = np.sin(phi_rad*(m+1))
            Pnm.append(sinth*Pnm_tmp / rootn[m+m+2])
            dPnm.append((dPnm_diag_tmp* rootn[m+1] * np.sqrt(.5)))#/(2*(m+1)+1))
            k = kstart + 1
            calc_sha(k,m+1,m+1,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
            Pnm.pop(0)
            dPnm.pop(0)


    mask = ~np.isnan(A_r).any(axis=1)
    A_r = A_r[mask]
    A_theta = A_theta[mask]
    A_phi = A_phi[mask]

    if calcTP:
        return A_r, A_theta, A_phi

    else:
        return A_r


    # ### Leg_SHA_optimized
def Leg_SHA_optimized(rho, theta, phi, nmax, mmax, get='all',epsilon = 1e-2, silence_warnings = False):
    """
    Written by Collin Kadlecek 7/2024, Collin.kadlecek@noaa.gov
    Computes matrices to connect the radial, colatitude and azimuthal field
    components to the magnetic potential field in terms of spherical harmonic
    coefficients (Schmidt quasi-normalized). In order major order
    i.e. g10,g20,g30...gnmax0, g11,h11,g21,h21...

    Parameters
    ----------

    radius : ndarray, shape (...)
        Array containing the radius in earth radii.
    theta : ndarray, shape (...)
        Array containing the colatitude in degrees
        :math:`[0^\\circ,180^\\circ]`.
    phi : ndarray, shape (...)
        Array containing the longitude in degrees.
    nmax : int, positive
        Maximum degree of the sphercial harmonic expansion.
    mmax : int, positive
        Maximum order of the spherical harmonic expansion (defaults to
        ``nmax``). For ``mmax = 0``, for example, only the zonal terms are
        returned.
    get : {'r', 'all'}, optional
        Return just the radial component of the field, or all components

    Returns
    -------
    A_radius, A_theta, A_phi : ndarray, shape (3, nmax*(nmax+2), len(theta))
        Matrices for radial, colatitude and azimuthal field components. The
        second dimension ``M`` varies depending on ``nmax``, ``nmin`` and
        ``mmax``.

    Algorithm description
    --------
    The implimentation of this function follows the structure of the associated
    legendre polynomial generation function (legendre_poly) created by Nils Olsen from the
    Chaos magpy package. This function recursively generates the associated legendre
    polynomials. Leg_SHA follows this recursive generations structure. However, it
    also generates each polynomials derivative for each n,m index. Further, only
    the necessary legendre polynomials are held in memory. Upon each polynomials
    generation, it's associated matrix coefficients are calculated.

    There are two recursion relations used for the polynomial generation. These
    are equations 26 and 27 from The Main GeoMagnetic Field Langel Chapter 4. 26 is
    used to generate diagonal (n == m) indeces, and 27 for all other.

    The derivative is generated with a recurrence relation from relation from Holmes and
    Featherstone. There are two recursion relations which follow the structures of the
    polynomial generation:
    NOTE:
    dP[l,k] refers to the derivative of legendre polynomial with degree = l and order = k
    (where degree >= order)
    costh = cos(theta) sinth = sin(theta)

    For off diagonal terms (m = n-1) which is a special case of 27:
    dP[n,n-1] = dP[n-1,n-1]*costh*sqrt(n+n+1)-sinth*P[n-1,n-1]*sqrt(n+n+1)

    For diagonal terms (m = n) following 26:
    dP[n,n] = P[n,n-1]*sqrt(n)/sqrt(2)

    For all other terms:
    dP[n,m] = (n*costh*P[n,m] - sqrt(n**2-m**2)*P[n-1,m])/sinth


    Warnings
    --------
    The function can also return the design matrix for the field components at
    the geographic poles, i.e. where ``theta == 0.`` or ``theta == 180.``.
    However, users should be careful when doing this because the vector basis
    for spherical geocentric coordinates,
    :math:`{{\\mathbf{e}_r, \\mathbf{e}_\\theta, \\mathbf{e}_\\phi}}`,
    depends on longitude, which is not well defined at the poles. That is,
    at the poles, any value for the longitude maps to the same location in
    euclidean coordinates but gives a different vector basis in spherical
    geocentric coordinates. Nonetheless, by choosing a specific value for the
    longitude at the poles, users can define the vector basis, which then
    establishes the meaning of the spherical geocentric components. The vector
    basis for the horizontal components is defined as

    .. math::

        \\mathbf{e}_\\theta &= \\cos\\theta\\cos\\phi\\mathbf{e}_x -
            \\cos\\theta\\sin\\phi\\mathbf{e}_y - \\sin\\theta\\mathbf{e}_z\\\\
        \\mathbf{e}_\\phi &= -\\sin\\phi\\mathbf{e}_x +
            \\cos\\phi\\mathbf{e}_y

    Hence, at the geographic north pole as given by ``theta = 0.`` and
    ``phi = 0.`` (chosen by the user), the returned design matrix ``A_theta``
    will be for components along the direction
    :math:`\\mathbf{e}_\\theta = \\mathbf{e}_x` and ``A_phi`` for components
    along :math:`\\mathbf{e}_\\phi = \\mathbf{e}_y`. However,
    if ``phi = 180.`` is chosen, ``A_theta`` will be for components along
    :math:`\\mathbf{e}_\\theta = -\\mathbf{e}_x` and ``A_phi``
    along :math:`\\mathbf{e}_\\phi = -\\mathbf{e}_y`.

    Examples
    --------
    Create the design matrices given 4 locations on the Earth's surface:

    >>> r = 1 = 6371.2km
    >>> theta = np.array([90., 100., 110., 120.])
    >>> phi = 0.
    >>> A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=1)
    >>> A_radius
    array([[ 1.22464680e-16, 2.00000000e+00, 0.00000000e+00],
           [-3.47296355e-01, 1.96961551e+00, 0.00000000e+00],
           [-6.84040287e-01, 1.87938524e+00, 0.00000000e+00],
           [-1.00000000e+00, 1.73205081e+00, 0.00000000e+00]])

    Say, only the axial dipole coefficient is non-zero, what is the magnetic
    field in terms of spherical geocentric components?

    >>> m = np.array([-30000, 0., 0.])  # model coefficients in nT
    >>> Br = A_radius @ m; Br
    array([-3.67394040e-12, 1.04188907e+04, 2.05212086e+04, 3.00000000e+04])
    >>> Bt = A_theta @ m; Bt
    array([-30000. , -29544.23259037, -28190.77862358, -25980.76211353])
    >>> Bp = A_phi @ m; Bp
    array([0., 0., 0., 0.])

    A more complete example is given below:

    .. code-block:: python

        import numpy as np
        from chaosmagpy.model_utils import design_gauss, synth_values

        nmax = 5  # desired truncation degree, i.e. 35 model coefficients
        coeffs = np.arange(35)  # example model coefficients

        # example locations
        N = 10
        radius = 1 = 6371.2 km # Earth's surface
        phi = np.linspace(-180., 180., num=N)  # azimuth in degrees
        theta = np.linspace(1., 179., num=N)  # colatitude in degrees

        # compute design matrices to compute predictions of the
        # field components from the model coefficients, each is of shape N x 35
        A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=nmax)

        # magnetic components computed from the model
        Br_pred = A_radius @ coeffs
        Bt_pred = A_theta @ coeffs
        Bp_pred = A_phi @ coeffs

        # compute the magnetic components directly from the coefficients
        Br, Bt, Bp = synth_values(coeffs, radius, theta, phi)
        np.linalg.norm(Br - Br_pred)  # approx 0.
        np.linalg.norm(Bt - Bt_pred)  # approx 0.
        np.linalg.norm(Bp - Bp_pred)  # approx 0.

    """

    #Locally define a function for calculating SHA
    def calc_sha(k,n,m,rn1,Pnm,dPnm, cos_phi_mem,sin_phi_mem, sin_theta):
        dPnm = -dPnm
        nr = (n+1) * rn1
        tr = rn1*inverse_sinth
        #g terms as in g11, g21, g22 etc...
        A_r[k] = nr*Pnm*cos_phi_mem
        A_theta[k] = rn1*dPnm*cos_phi_mem
        A_phi[k] = tr*Pnm*(m*sin_phi_mem)
        k+=1

        if(m!=0):

            #h terms as in h11, h21, h22 etc...
            A_r[k] = nr*Pnm*sin_phi_mem
            A_theta[k] = rn1*dPnm*sin_phi_mem
            A_phi[k] = -tr*Pnm*m*cos_phi_mem
            k = k + 1
        # else:
        #     A_r[k] = np.nan + np.zeros(len(costh))
        #     A_theta[k] = np.nan + np.zeros(len(costh))
        #     A_phi[k] = np.nan + np.zeros(len(costh))
        return k


    theta_min = np.amin(theta)
    theta_max = np.amax(theta)

    # check if poles are included
    if (np.isclose(0,theta_min,atol = epsilon)) or (np.isclose(theta_max,180,atol = epsilon)):
        if(not silence_warnings):
            strtemp = f'Input coordinates include the poles. They have been shifted by {epsilon}'
            warnings.warn(strtemp, stacklevel=2)
        mask_0 = np.isclose(theta, 0, atol=epsilon)
        mask_180 = np.isclose(theta, 180, atol=epsilon)
        theta[mask_0 | mask_180] = epsilon
        poles = True
    if (theta_min < 0.0) or (theta_max > 180.0):
        raise ValueError('Colatitude outside bounds [0, 180].')

    if(not silence_warnings):
        if (np.isclose(theta_max, 3.14, atol= .01)):
            warnings.warn('Please ensure that your inputs are in earth radii (6371.2km)\n and degrees for both longitude and colatitude')

    if get=='r':
        calcTP=False
    else:
        calcTP=True
    N_data=np.size(theta,0)
    if (np.size(phi,0) != N_data or np.size(rho,0) != N_data):
        raise ValueError(r"Leg_SHA_optimized: rho, theta, and phi have different dimensions")
    if np.isscalar(theta) is not True and theta.ndim > 1:
        dim2=np.size(theta,1)
    else:
        dim2 = 1
    if dim2 > 1:
        theta=theta.transpose()
        phi=phi.transpose()

    costh = np.cos(np.deg2rad(theta))
    sinth = np.sqrt(1-costh*costh)
    inverse_sinth = 1/sinth
    if np.isscalar(rho) is True:
        rho=rho*np.ones(N_data)
    # N_nm=mmax*(mmax + 2) + (nmax + mmax)*(2*mmax + 1)
    N_nm=mmax*(mmax + 2)# (nmax - mmax)*(2*mmax + 1)
    # calculate sub-matrices for s=0, p=0
    #This looks strange, but np.zeros takes a tuple as input, thus the extra paren is necessary
    A_r=np.zeros((N_nm,N_data))
    if(calcTP):
        A_theta=np.zeros((N_nm,N_data))
        A_phi=np.zeros((N_nm,N_data))

    rootn = np.sqrt(np.arange(2 * nmax**2 + 1))
    # Compute the power array
    powers = -(np.arange(nmax) + 3)

    # Raise each element in rho to each power
    r2n = np.transpose(rho[:, np.newaxis] ** powers)
    phi_rad = np.deg2rad(phi)
    Pnm = []
    dPnm = []

    #This is done twice to begin the Pnm queue
    Pnm.append(np.ones(len(costh)))  # is copied into trailing dimensions
    dPnm.append(np.zeros(len(costh)))
    Pnm.append(np.ones(len(costh)))
    dPnm.append(np.zeros(len(costh)))
    k=0 #Cycles through all the coefficients for degree n, order m model.

    # k = calc_sha(k,0,0,r2n[0],1,0,np.cos(phi_rad),np.sin(phi_rad),sinth)



    cos_phi = np.ones(len(phi))
    sin_phi = cos_phi-1
    for m in range(nmax):


        if(m == 1):

            Pnm.append(sinth)
            dPnm.append(costh)
            k = calc_sha(k,1,1,r2n[0],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)


        # Buffer normalization factor for [n-1,n-1] to [n,n-1]
        c2 = rootn[m + m + 1]
        #Pnm_tmp holds Pnm[m,m] with normalization factor
        Pnm_tmp = c2 * Pnm[-1]
        #Pnm[m,m-1] = Pnm[m-1,m-1]*cos*c2
        Pnm.append(costh * Pnm_tmp)

        #Hold this for the derivative of diagonal later
        dPnm_diag_tmp = Pnm[-1]
        #Derivative of previous diagonal * cos - sin * previous diagonal
        dPnm.append(dPnm[-1]*costh*c2 - sinth * Pnm_tmp)
        k = calc_sha(k,m+1,m,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
        Pnm.pop(0)
        dPnm.pop(0)

        for n in range(m+2, nmax+1):
            d = n * n - m * m
            e = n + n - 1


            Pnm.append((e * costh *  Pnm[-1] - rootn[d-e] * Pnm[-2])
                         / rootn[d])
            #dPnm[n,m] = (n*cos*Pnm[n,m] - d^1/2 * Pnm[n-1,m])/sin**2

            dPnm.append((n*costh*Pnm[-1] - rootn[d]* Pnm[-2])*inverse_sinth)
            k = calc_sha(k,n,m,r2n[n-1],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
            Pnm.pop(0)
            dPnm.pop(0)

        cos_phi = np.cos(phi_rad*(m+1))
        sin_phi = np.sin(phi_rad*(m+1))
        if m > 0:#Diagonal append Pnm[m,m] = sin^m(theta)

            Pnm.append(sinth*Pnm_tmp / rootn[m+m+2])
            dPnm.append((dPnm_diag_tmp* rootn[m+1] * np.sqrt(.5)))

            k = calc_sha(k,m+1,m+1,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)

            Pnm.pop(0)
            dPnm.pop(0)

    # dP(n,m) = Pnm(m,n+1) is the derivative of P(n,m) vrt. theta
    if calcTP:
        return A_r, A_theta, A_phi
    else:
        return A_r

def Leg_SHA_optimized_with_nmin(rho, theta, phi, nmin, nmax, get='all',epsilon = 1e-2, silence_warnings = False):
    """
    Written by Collin Kadlecek 7/2024, Collin.kadlecek@noaa.gov
    Computes matrices to connect the radial, colatitude and azimuthal field
    components to the magnetic potential field in terms of spherical harmonic
    coefficients (Schmidt quasi-normalized). In order major order
    i.e. g10,g20,g30...gnmax0, g11,h11,g21,h21...

    Parameters
    ----------

    radius : ndarray, shape (...)
        Array containing the radius in earth radii.
    theta : ndarray, shape (...)
        Array containing the colatitude in degrees
        :math:`[0^\\circ,180^\\circ]`.
    phi : ndarray, shape (...)
        Array containing the longitude in degrees.
    nmax : int, positive
        Maximum degree of the sphercial harmonic expansion.
    mmax : int, positive
        Maximum order of the spherical harmonic expansion (defaults to
        ``nmax``). For ``mmax = 0``, for example, only the zonal terms are
        returned.
    get : {'r', 'all'}, optional
        Return just the radial component of the field, or all components

    Returns
    -------
    A_radius, A_theta, A_phi : ndarray, shape (3, nmax*(nmax+2), len(theta))
        Matrices for radial, colatitude and azimuthal field components. The
        second dimension ``M`` varies depending on ``nmax``, ``nmin`` and
        ``mmax``.

    Algorithm description
    --------
    The implimentation of this function follows the structure of the associated
    legendre polynomial generation function (legendre_poly) created by Nils Olsen from the
    Chaos magpy package. This function recursively generates the associated legendre
    polynomials. Leg_SHA follows this recursive generations structure. However, it
    also generates each polynomials derivative for each n,m index. Further, only
    the necessary legendre polynomials are held in memory. Upon each polynomials
    generation, it's associated matrix coefficients are calculated.

    There are two recursion relations used for the polynomial generation. These
    are equations 26 and 27 from The Main GeoMagnetic Field Langel Chapter 4. 26 is
    used to generate diagonal (n == m) indeces, and 27 for all other.

    The derivative is generated with a recurrence relation from relation from Holmes and
    Featherstone. There are two recursion relations which follow the structures of the
    polynomial generation:
    NOTE:
    dP[l,k] refers to the derivative of legendre polynomial with degree = l and order = k
    (where degree >= order)
    costh = cos(theta) sinth = sin(theta)

    For off diagonal terms (m = n-1) which is a special case of 27:
    dP[n,n-1] = dP[n-1,n-1]*costh*sqrt(n+n+1)-sinth*P[n-1,n-1]*sqrt(n+n+1)

    For diagonal terms (m = n) following 26:
    dP[n,n] = P[n,n-1]*sqrt(n)/sqrt(2)

    For all other terms:
    dP[n,m] = (n*costh*P[n,m] - sqrt(n**2-m**2)*P[n-1,m])/sinth


    Warnings
    --------
    The function can also return the design matrix for the field components at
    the geographic poles, i.e. where ``theta == 0.`` or ``theta == 180.``.
    However, users should be careful when doing this because the vector basis
    for spherical geocentric coordinates,
    :math:`{{\\mathbf{e}_r, \\mathbf{e}_\\theta, \\mathbf{e}_\\phi}}`,
    depends on longitude, which is not well defined at the poles. That is,
    at the poles, any value for the longitude maps to the same location in
    euclidean coordinates but gives a different vector basis in spherical
    geocentric coordinates. Nonetheless, by choosing a specific value for the
    longitude at the poles, users can define the vector basis, which then
    establishes the meaning of the spherical geocentric components. The vector
    basis for the horizontal components is defined as

    .. math::

        \\mathbf{e}_\\theta &= \\cos\\theta\\cos\\phi\\mathbf{e}_x -
            \\cos\\theta\\sin\\phi\\mathbf{e}_y - \\sin\\theta\\mathbf{e}_z\\\\
        \\mathbf{e}_\\phi &= -\\sin\\phi\\mathbf{e}_x +
            \\cos\\phi\\mathbf{e}_y

    Hence, at the geographic north pole as given by ``theta = 0.`` and
    ``phi = 0.`` (chosen by the user), the returned design matrix ``A_theta``
    will be for components along the direction
    :math:`\\mathbf{e}_\\theta = \\mathbf{e}_x` and ``A_phi`` for components
    along :math:`\\mathbf{e}_\\phi = \\mathbf{e}_y`. However,
    if ``phi = 180.`` is chosen, ``A_theta`` will be for components along
    :math:`\\mathbf{e}_\\theta = -\\mathbf{e}_x` and ``A_phi``
    along :math:`\\mathbf{e}_\\phi = -\\mathbf{e}_y`.

    Examples
    --------
    Create the design matrices given 4 locations on the Earth's surface:

    >>> r = 1 = 6371.2km
    >>> theta = np.array([90., 100., 110., 120.])
    >>> phi = 0.
    >>> A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=1)
    >>> A_radius
    array([[ 1.22464680e-16, 2.00000000e+00, 0.00000000e+00],
           [-3.47296355e-01, 1.96961551e+00, 0.00000000e+00],
           [-6.84040287e-01, 1.87938524e+00, 0.00000000e+00],
           [-1.00000000e+00, 1.73205081e+00, 0.00000000e+00]])

    Say, only the axial dipole coefficient is non-zero, what is the magnetic
    field in terms of spherical geocentric components?

    >>> m = np.array([-30000, 0., 0.])  # model coefficients in nT
    >>> Br = A_radius @ m; Br
    array([-3.67394040e-12, 1.04188907e+04, 2.05212086e+04, 3.00000000e+04])
    >>> Bt = A_theta @ m; Bt
    array([-30000. , -29544.23259037, -28190.77862358, -25980.76211353])
    >>> Bp = A_phi @ m; Bp
    array([0., 0., 0., 0.])

    A more complete example is given below:

    .. code-block:: python

        import numpy as np
        from chaosmagpy.model_utils import design_gauss, synth_values

        nmax = 5  # desired truncation degree, i.e. 35 model coefficients
        coeffs = np.arange(35)  # example model coefficients

        # example locations
        N = 10
        radius = 1 = 6371.2 km # Earth's surface
        phi = np.linspace(-180., 180., num=N)  # azimuth in degrees
        theta = np.linspace(1., 179., num=N)  # colatitude in degrees

        # compute design matrices to compute predictions of the
        # field components from the model coefficients, each is of shape N x 35
        A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=nmax)

        # magnetic components computed from the model
        Br_pred = A_radius @ coeffs
        Bt_pred = A_theta @ coeffs
        Bp_pred = A_phi @ coeffs

        # compute the magnetic components directly from the coefficients
        Br, Bt, Bp = synth_values(coeffs, radius, theta, phi)
        np.linalg.norm(Br - Br_pred)  # approx 0.
        np.linalg.norm(Bt - Bt_pred)  # approx 0.
        np.linalg.norm(Bp - Bp_pred)  # approx 0.

    """
    if(nmin == 0):
        nmin = -1
    #Locally define a function for calculating SHA
    def calc_sha(k,n,m,rn1,Pnm,dPnm, cos_phi_mem,sin_phi_mem, sin_theta):
        if(n < nmin):
            return k
        dPnm = -dPnm
        nr = (n+1) * rn1
        tr = rn1*inverse_sinth
        #g terms as in g11, g21, g22 etc...
        A_r[k] = nr*Pnm*cos_phi_mem
        A_theta[k] = rn1*dPnm*cos_phi_mem
        A_phi[k] = tr*Pnm*(m*sin_phi_mem)
        k+=1

        if(m!=0):

            #h terms as in h11, h21, h22 etc...
            A_r[k] = nr*Pnm*sin_phi_mem
            A_theta[k] = rn1*dPnm*sin_phi_mem
            A_phi[k] = -tr*Pnm*m*cos_phi_mem
            k = k + 1
        # else:
        #     A_r[k] = np.nan + np.zeros(len(costh))
        #     A_theta[k] = np.nan + np.zeros(len(costh))
        #     A_phi[k] = np.nan + np.zeros(len(costh))
        return k


    theta_min = np.amin(theta)
    theta_max = np.amax(theta)

    # check if poles are included
    if (np.isclose(0,theta_min,atol = epsilon)) or (np.isclose(theta_max,180,atol = epsilon)):
        if(not silence_warnings):
            strtemp = f'Input coordinates include the poles. They have been shifted by {epsilon}'
            warnings.warn(strtemp, stacklevel=2)
        mask_0 = np.isclose(theta, 0, atol=epsilon)
        mask_180 = np.isclose(theta, 180, atol=epsilon)
        theta[mask_0 | mask_180] = epsilon
        poles = True
    if (theta_min < 0.0) or (theta_max > 180.0):
        raise ValueError('Colatitude outside bounds [0, 180].')

    if(not silence_warnings):
        if (np.isclose(theta_max, 3.14, atol= .01)):
            warnings.warn('Please ensure that your inputs are in earth radii (6371.2km) and degrees for both longitude and colatitude')

    if get=='r':
        calcTP=False
    else:
        calcTP=True
    N_data=np.size(theta,0)
    if (np.size(phi,0) != N_data or np.size(rho,0) != N_data):
        raise ValueError(r"Leg_SHA_optimized: rho, theta, and phi have different dimensions")
    if np.isscalar(theta) is not True and theta.ndim > 1:
        dim2=np.size(theta,1)
    else:
        dim2 = 1
    if dim2 > 1:
        theta=theta.transpose()
        phi=phi.transpose()

    costh = np.cos(np.deg2rad(theta))
    sinth = np.sqrt(1-costh*costh)
    inverse_sinth = 1/sinth
    if np.isscalar(rho) is True:
        rho=rho*np.ones(N_data)
    # N_nm=mmax*(mmax + 2) + (nmax + mmax)*(2*mmax + 1)
    N_nm=nmax*(nmax + 2) - (nmin+1)*(nmin-1)
    # calculate sub-matrices for s=0, p=0
    #This looks strange, but np.zeros takes a tuple as input, thus the extra paren is necessary
    A_r=np.zeros((N_nm,N_data))
    if(calcTP):
        A_theta=np.zeros((N_nm,N_data))
        A_phi=np.zeros((N_nm,N_data))

    rootn = np.sqrt(np.arange(2 * nmax**2 + 1))
    # Compute the power array
    powers = -(np.arange(nmax) + 3)

    # Raise each element in rho to each power
    r2n = np.transpose(rho[:, np.newaxis] ** powers)
    phi_rad = np.deg2rad(phi)
    Pnm = []
    dPnm = []

    #This is done twice to begin the Pnm queue
    Pnm.append(np.ones(len(costh)))  # is copied into trailing dimensions
    dPnm.append(np.zeros(len(costh)))
    Pnm.append(np.ones(len(costh)))
    dPnm.append(np.zeros(len(costh)))
    k=0 #Cycles through all the coefficients for degree n, order m model.

    # k = calc_sha(k,0,0,r2n[0],1,0,np.cos(phi_rad),np.sin(phi_rad),sinth)



    cos_phi = np.ones(len(phi))
    sin_phi = cos_phi-1
    for m in range(nmax):


        if(m == 1):

            Pnm.append(sinth)
            dPnm.append(costh)
            k = calc_sha(k,1,1,r2n[0],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)


        # Buffer normalization factor for [n-1,n-1] to [n,n-1]
        c2 = rootn[m + m + 1]
        #Pnm_tmp holds Pnm[m,m] with normalization factor
        Pnm_tmp = c2 * Pnm[-1]
        #Pnm[m,m-1] = Pnm[m-1,m-1]*cos*c2
        Pnm.append(costh * Pnm_tmp)

        #Hold this for the derivative of diagonal later
        dPnm_diag_tmp = Pnm[-1]
        #Derivative of previous diagonal * cos - sin * previous diagonal
        dPnm.append(dPnm[-1]*costh*c2 - sinth * Pnm_tmp)
        k = calc_sha(k,m+1,m,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
        Pnm.pop(0)
        dPnm.pop(0)

        for n in range(m+2, nmax+1):
            d = n * n - m * m
            e = n + n - 1


            Pnm.append((e * costh *  Pnm[-1] - rootn[d-e] * Pnm[-2])
                         / rootn[d])
            #dPnm[n,m] = (n*cos*Pnm[n,m] - d^1/2 * Pnm[n-1,m])/sin**2

            dPnm.append((n*costh*Pnm[-1] - rootn[d]* Pnm[-2])*inverse_sinth)
            k = calc_sha(k,n,m,r2n[n-1],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
            Pnm.pop(0)
            dPnm.pop(0)

        cos_phi = np.cos(phi_rad*(m+1))
        sin_phi = np.sin(phi_rad*(m+1))
        if m > 0:#Diagonal append Pnm[m,m] = sin^m(theta)

            Pnm.append(sinth*Pnm_tmp / rootn[m+m+2])
            dPnm.append((dPnm_diag_tmp* rootn[m+1] * np.sqrt(.5)))

            k = calc_sha(k,m+1,m+1,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)

            Pnm.pop(0)
            dPnm.pop(0)

    # dP(n,m) = Pnm(m,n+1) is the derivative of P(n,m) vrt. theta
    if calcTP:
        return A_r, A_theta, A_phi
    else:
        return A_r


def Leg_SHA_optimized_ind_nmin(rho, theta, phi, nmin, nmax, get='all', epsilon = 1e-2, silence_warnings = False):
    """
    Written by Collin Kadlecek 7/2024 Collin.kadlecek@noaa.gov
    Computes matrices to connect the radial, colatitude and azimuthal field
    components to the magnetic potential field in terms of spherical harmonic
    coefficients (Schmidt quasi-normalized). In degree order 
    i.e. g10,g20,g30...gnmax0, g11,h11,g21,h21...

    Parameters
    ----------

    radius : ndarray, shape (...)
        Array containing the radius in earth radii.
    theta : ndarray, shape (...)
        Array containing the colatitude in degrees
        :math:`[0^\\circ,180^\\circ]`.
    phi : ndarray, shape (...)
        Array containing the longitude in degrees.
    nmax : int, positive
        Maximum degree of the sphercial harmonic expansion.
    mmax : int, positive
        Maximum order of the spherical harmonic expansion (defaults to
        ``nmax``). For ``mmax = 0``, for example, only the zonal terms are
        returned.
    get : {'r', 'all'}, optional
        Return just the radial component of the field, or all components

    Returns
    -------
    A_radius, A_theta, A_phi : ndarray, shape (3, nmax*(nmax+2), len(theta))
        Matrices for radial, colatitude and azimuthal field components. The
        second dimension ``M`` varies depending on ``nmax``, ``nmin`` and
        ``mmax``.

    Algorithm description
    --------
    The implimentation of this function follows the structure of the associated
    legendre polynomial generation function (legendre_poly) created by Nils Olsen from the 
    Chaos magpy package. This function recursively generates the associated legendre
    polynomials. Leg_SHA follows this recursive generations structure. However, it 
    also generates each polynomials derivative for each n,m index. Further, only
    the necessary legendre polynomials are held in memory. Upon each polynomials
    generation, it's associated matrix coefficients are calculated. 

    There are two recursion relations used for the polynomial generation. These
    are equations 26 and 27 from The Main GeoMagnetic Field Langel Chapter 4. 26 is 
    used to generate diagonal (n == m) indeces, and 27 for all other. 

    A different source is used to generate the derivatives. There are two recursion 
    relations which follow the structures of the polynomial generation:
    NOTE:
    dP[l,k] refers to the derivative of legendre polynomial with degree = l and order = k
    (where degree >= order)
    costh = cos(theta) sinth = sin(theta)

    For off diagonal terms (m = n-1) which is a special case of 27:
    dP[n,n-1] = dP[n-1,n-1]*costh*sqrt(n+n+1)-sinth*P[n-1,n-1]*sqrt(n+n+1)
    
    For diagonal terms (m = n) following 26:
    dP[n,n] = P[n,n-1]*sqrt(n)/sqrt(2)

    For all other terms:
    dP[n,m] = (n*costh*P[n,m] - sqrt(n**2-m**2)*P[n-1,m])/sinth
    

    Warnings
    --------
    The function can also return the design matrix for the field components at
    the geographic poles, i.e. where ``theta == 0.`` or ``theta == 180.``.
    However, users should be careful when doing this because the vector basis
    for spherical geocentric coordinates,
    :math:`{{\\mathbf{e}_r, \\mathbf{e}_\\theta, \\mathbf{e}_\\phi}}`,
    depends on longitude, which is not well defined at the poles. That is,
    at the poles, any value for the longitude maps to the same location in
    euclidean coordinates but gives a different vector basis in spherical
    geocentric coordinates. Nonetheless, by choosing a specific value for the
    longitude at the poles, users can define the vector basis, which then
    establishes the meaning of the spherical geocentric components. The vector
    basis for the horizontal components is defined as

    .. math::

        \\mathbf{e}_\\theta &= \\cos\\theta\\cos\\phi\\mathbf{e}_x -
            \\cos\\theta\\sin\\phi\\mathbf{e}_y - \\sin\\theta\\mathbf{e}_z\\\\
        \\mathbf{e}_\\phi &= -\\sin\\phi\\mathbf{e}_x +
            \\cos\\phi\\mathbf{e}_y

    Hence, at the geographic north pole as given by ``theta = 0.`` and
    ``phi = 0.`` (chosen by the user), the returned design matrix ``A_theta``
    will be for components along the direction
    :math:`\\mathbf{e}_\\theta = \\mathbf{e}_x` and ``A_phi`` for components
    along :math:`\\mathbf{e}_\\phi = \\mathbf{e}_y`. However,
    if ``phi = 180.`` is chosen, ``A_theta`` will be for components along
    :math:`\\mathbf{e}_\\theta = -\\mathbf{e}_x` and ``A_phi``
    along :math:`\\mathbf{e}_\\phi = -\\mathbf{e}_y`.

    Examples
    --------
    Create the design matrices given 4 locations on the Earth's surface:

    >>> r = 1 = 6371.2km
    >>> theta = np.array([90., 100., 110., 120.])
    >>> phi = 0.
    >>> A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=1)
    >>> A_radius
    array([[ 1.22464680e-16, 2.00000000e+00, 0.00000000e+00],
           [-3.47296355e-01, 1.96961551e+00, 0.00000000e+00],
           [-6.84040287e-01, 1.87938524e+00, 0.00000000e+00],
           [-1.00000000e+00, 1.73205081e+00, 0.00000000e+00]])

    Say, only the axial dipole coefficient is non-zero, what is the magnetic
    field in terms of spherical geocentric components?

    >>> m = np.array([-30000, 0., 0.])  # model coefficients in nT
    >>> Br = A_radius @ m; Br
    array([-3.67394040e-12, 1.04188907e+04, 2.05212086e+04, 3.00000000e+04])
    >>> Bt = A_theta @ m; Bt
    array([-30000. , -29544.23259037, -28190.77862358, -25980.76211353])
    >>> Bp = A_phi @ m; Bp
    array([0., 0., 0., 0.])

    A more complete example is given below:

    .. code-block:: python

        import numpy as np
        from chaosmagpy.model_utils import design_gauss, synth_values

        nmax = 5  # desired truncation degree, i.e. 35 model coefficients
        coeffs = np.arange(35)  # example model coefficients

        # example locations
        N = 10
        radius = 1 = 6371.2 km # Earth's surface
        phi = np.linspace(-180., 180., num=N)  # azimuth in degrees
        theta = np.linspace(1., 179., num=N)  # colatitude in degrees

        # compute design matrices to compute predictions of the
        # field components from the model coefficients, each is of shape N x 35
        A_radius, A_theta, A_phi = design_gauss(r, theta, phi, nmax=nmax)

        # magnetic components computed from the model
        Br_pred = A_radius @ coeffs
        Bt_pred = A_theta @ coeffs
        Bp_pred = A_phi @ coeffs

        # compute the magnetic components directly from the coefficients
        Br, Bt, Bp = synth_values(coeffs, radius, theta, phi)
        np.linalg.norm(Br - Br_pred)  # approx 0.
        np.linalg.norm(Bt - Bt_pred)  # approx 0.
        np.linalg.norm(Bp - Bp_pred)  # approx 0.

    """
    del_indexes = []
    if(nmin == 0):
        nmin = nmin-1
    def calc_sha(k,n,m,rn1,Pnm,dPnm, cos_phi_mem,sin_phi_mem, sin_theta):
        k = k - 1
        k = 2*k
        if(n < nmin):
            del_indexes.append(k+1)
            A_r[k] = np.nan + np.zeros(len(costh))
            A_theta[k] = np.nan + np.zeros(len(costh))
            A_phi[k] = np.nan + np.zeros(len(costh))

            A_r[k +1] = np.nan + np.zeros(len(costh))
            A_theta[k +1] = np.nan + np.zeros(len(costh))
            A_phi[k +1] = np.nan + np.zeros(len(costh))
            return
        dPnm = -dPnm
        nr = (n+1) * rn1
        tr = rn1*inverse_sinth
        A_r[k] = nr*Pnm*cos_phi_mem
        A_theta[k] = rn1*dPnm*cos_phi_mem
        A_phi[k] = -tr*Pnm*(-m*sin_phi_mem)
        if(m!=0):
            k+=1
            #h terms as in h11, h21, h22 etc...
            A_r[k] = nr*Pnm*sin_phi_mem
            A_theta[k] = rn1*dPnm*sin_phi_mem
            A_phi[k] = -tr*Pnm*m*cos_phi_mem
        else:
            del_indexes.append(k+1)
            A_r[k +1] = np.nan + np.zeros(len(costh))
            A_theta[k +1] = np.nan + np.zeros(len(costh))
            A_phi[k +1] = np.nan + np.zeros(len(costh))

    if get=='r':
        calcTP=False
    else:
        calcTP=True
    N_data=np.size(theta,0)

    theta_min = np.amin(theta)
    theta_max = np.amax(theta)

    # check if poles are included
    if (np.isclose(0,theta_min,epsilon)) or (np.isclose(theta_max,180,epsilon)):
        if(not silence_warnings):
            warnings.warn(f'Input coordinates include the poles. They have been shifted by {epsilon}')
        mask_0 = np.isclose(theta, 0, atol=epsilon)
        mask_180 = np.isclose(theta, 180, atol=epsilon)
        theta[mask_0 | mask_180] = epsilon
        poles = True
    if (theta_min <= 0.0) or (theta_max >= 180.0):
        raise ValueError('Colatitude outside bounds [0, 180].')
    if (np.isclose(theta_max, 3.14, atol= .01)):
        warnings.warn('Please ensure that your inputs are in earth radii (6371.2km) and degrees for both longitude and colatitude')

    if np.isscalar(theta) is not True and theta.ndim > 1:
        dim2=np.size(theta,1)
    else:
        dim2 = 1
    if dim2 > 1:
        theta=theta.transpose()
        phi=phi.transpose()

    costh = np.cos(np.deg2rad(theta))
    sinth = np.sqrt(1-costh*costh)


    inverse_sinth = 1/sinth

    if np.isscalar(rho) is True:
        rho=rho*np.ones(N_data)
    N_nm=nmax*(nmax + 2) + nmax
    #This looks strange, but np.zeros takes a tuple as input, thus the extra paren is necessary
    A_r=np.zeros((N_nm,N_data))
    if(calcTP):
        A_theta=np.zeros((N_nm,N_data))
        A_phi=np.zeros((N_nm,N_data))

    rootn = np.sqrt(np.arange(2 * nmax**2 + 1))
    # Compute the power array
    powers = -(np.arange(nmax) + 3)

    # Raise each element in rho to each power
    r2n = np.transpose(rho[:, np.newaxis] ** powers)
    # Recursion relations after Langel "The Main Field" (1987),
    # eq. (27) and Table 2 (p. 256)
    phi_rad = np.deg2rad(phi)
    Pnm = []
    dPnm = []


    Pnm.append(np.ones(len(costh)))  # is copied into trailing dimensions
    dPnm.append(np.zeros(len(costh)))
    Pnm.append(np.ones(len(costh)))  # is copied into trailing dimensions
    dPnm.append(np.zeros(len(costh)))
    k=0 #Cycles through all the coefficients for degree n, order m model.


    for m in range(nmax):

        cos_phi = np.cos(phi_rad*m)
        sin_phi = np.sin(phi_rad*m)
        if(m == 1):
            Pnm.append(sinth)
            dPnm.append(costh)
            k = 2
            calc_sha(k,1,1,r2n[0],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)


        # Buffer normalization factor for [n-1,n-1] to [n,n-1]
        c2 = rootn[m + m + 1]
        #Pnm_tmp holds Pnm[m,m] with normalization factor
        Pnm_tmp = c2 * Pnm[-1]
        Pnm.append(costh * Pnm_tmp)

        #Hold this for the derivative of diagonal later
        dPnm_diag_tmp = Pnm[-1]
        #Derivative of previous diagonal * cos - sin * previous diagonal
        dPnm.append(dPnm[-1]*costh*c2 - sinth * Pnm_tmp)
        kstart = int((m+1)*(m+2)/2+m)
        calc_sha(kstart,m+1,m,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)

        Pnm.pop(0)
        dPnm.pop(0)
        k = kstart
        for n in range(m+2, nmax+1):
            d = n * n - m * m
            e = n + n - 1
            k = k + n

            Pnm.append((e * costh *  Pnm[-1] - rootn[d-e] * Pnm[-2])
                         / rootn[d])
            #Must be after Pnm is set to be setting derivative with same n
            #dPnm[n,m] = (n*cos*Pnm[n,m] - d^1/2 * Pnm[n-1,m])/sin**2

            dPnm.append((n*costh*Pnm[-1] - rootn[d]* Pnm[-2])*inverse_sinth)
            calc_sha(k,n,m,r2n[n-1],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
            Pnm.pop(0)
            dPnm.pop(0)

        if m > 0:#Diagonal append Pnm[m,m] = sin^m(theta)
            cos_phi = np.cos(phi_rad*(m+1))
            sin_phi = np.sin(phi_rad*(m+1))
            Pnm.append(sinth*Pnm_tmp / rootn[m+m+2])
            dPnm.append((dPnm_diag_tmp* rootn[m+1] * np.sqrt(.5)))#/(2*(m+1)+1))
            k = kstart + 1
            calc_sha(k,m+1,m+1,r2n[m],Pnm[-1],dPnm[-1],cos_phi,sin_phi,sinth)
            Pnm.pop(0)
            dPnm.pop(0)


    mask = ~np.isnan(A_r).any(axis=1)
    A_r = A_r[mask]
    A_theta = A_theta[mask]
    A_phi = A_phi[mask]

    if calcTP:
        return A_r, A_theta, A_phi

    else:
        return A_r

def permute_SHA(nmin, nmax):
    if(nmin != 0):
        N_nm = nmax*(nmax + 2)- (nmin+1)*(nmin-1)
    if(nmin == 0):
        N_nm = nmax*(nmax + 2)
    # print('total range = ', N_nm)
    permute = np.zeros((N_nm,N_nm))
    i = 0
    for m in range(0,nmax):

        kstart = int((m+1)*(m+2)/2+m)
        if(m+1 > nmin):
            permute[i,kstart] = 1
        i += 1
        k = kstart
        for n in range(m+2,nmax+1):
            k = k + n
            permute[i,k] = 1
            i += 1
        permute[i,kstart + 1] = 1
        i += 1
    return(permute)


def legendre_manoj(lat, nMax):
    """
    Computes  all of the Schmidt-semi normalized associated Legendre
    functions up to degree nMax. If nMax <= 16, function MAG_PcupLow is used.
    Otherwise MAG_PcupHigh is called.
    INPUT  Latitude (degrees)
    nMax        	integer 	 ( Maxumum degree of spherical harmonic secular model)
                    LegendreFunction Pointer to data structure with the following elements
                                                    double *Pcup;  (  pointer to store Legendre Function  )
                                                    double *dPcup; ( pointer to store  Derivative of Legendre function )
    LegendreFunction  Calculated Legendre variables in the data structure
    """

    sin_phi = np.sin(np.radians(lat))

    #    if nMax <= 16 or (1-abs(sin_phi)) < 10**-10:
    [P, dP] = MAG_PcupLow(sin_phi, nMax)
    # hoping that overflow doesn't occur
    #    else:
    #        raise ValueError("Implement Pcup High for this", nMax)

    return [P, dP]


def MAG_PcupLow(x, nMax):
    """ This function evaluates all of the Schmidt-semi normalized associated Legendre
        functions up to degree nMax.
        Calling Parameters:
                INPUT
                        nMax:	 Maximum spherical harmonic degree to compute.
                        x:		cos(colatitude) or sin(latitude).
                OUTPUT
                        Pcup:	A vector of all associated Legendgre polynomials evaluated at
                                        x up to nMax.
                   dPcup: Derivative of Pcup(x) with respect to latitude
        Notes: Overflow may occur if nMax > 20 , especially for high-latitudes.
        Use MAG_PcupHigh for large nMax.
    Written by Manoj Nair, June, 2009 . Manoj.C.Nair@Noaa.Gov.
    Note: In geomagnetism, the derivatives of ALF are usually found with
    respect to the colatitudes. Here the derivatives are found with respect
    to the latitude. The difference is a sign reversal for the derivative of
    the Associated Legendre Functions.
    """

    NumTerms = ((nMax + 1) * (nMax + 2) // 2)
    try:
        Pcup = np.zeros((NumTerms, x.size))
        dPcup = np.zeros((NumTerms, x.size))
    except AttributeError:

        Pcup = np.zeros((NumTerms, 1))
        dPcup = np.zeros((NumTerms, 1))
    Pcup[0] = 1.0
    dPcup[0] = 0.0
    # sin (geocentric latitude) - sin_phi
    z = np.sqrt((1.0 - x) * (1.0 + x))
    schmidtQuasiNorm = np.zeros(NumTerms)

    # schmidtQuasiNorm = (double *) malloc((NumTerms + 1) * sizeof ( double))

    # First, Compute the Gauss-normalized associated Legendre functions
    for n in range(1, nMax + 1):
        for m in range(n + 1):
            index = n * (n + 1) // 2 + m
            if n == m:
                index1 = (n - 1) * n // 2 + m - 1;
                Pcup[index] = z * Pcup[index1];
                dPcup[index] = z * dPcup[index1] + x * Pcup[index1];
            elif n == 1 and m == 0:
                index1 = (n - 1) * n // 2 + m;
                Pcup[index] = x * Pcup[index1];
                dPcup[index] = x * dPcup[index1] - z * Pcup[index1];
            elif n > 1 and n != m:
                index1 = (n - 2) * (n - 1) // 2 + m;
                index2 = (n - 1) * n // 2 + m;
                if m > n - 2:
                    Pcup[index] = x * Pcup[index2]
                    dPcup[index] = x * dPcup[index2] - z * Pcup[index2]
                else:
                    k = float(((n - 1) * (n - 1)) - (m * m)) / float((2 * n - 1) * (2 * n - 3))
                    Pcup[index] = x * Pcup[index2] - k * Pcup[index1];
                    dPcup[index] = x * dPcup[index2] - z * Pcup[index2] - k * dPcup[index1];

    # Compute the ratio between the the Schmidt quasi-normalized associated Legendre
    # functions and the Gauss-normalized version.

    schmidtQuasiNorm[0] = 1.0;
    for n in range(1, nMax + 1):
        index = n * (n + 1) // 2
        index1 = (n - 1) * n // 2
        # for m = 0
        schmidtQuasiNorm[index] = schmidtQuasiNorm[index1] * float((2 * n - 1)) / float(n)
        for m in range(1, n + 1):
            index = (n * (n + 1) // 2 + m)
            index1 = (n * (n + 1) // 2 + m - 1)
            schmidtQuasiNorm[index] = schmidtQuasiNorm[index1] * np.sqrt(
                float((n - m + 1) * (2 if m == 1 else 1)) / float(n + m))

    # Converts the  Gauss-normalized associated Legendre
    #          functions to the Schmidt quasi-normalized version using pre-computed
    #          relation stored in the variable schmidtQuasiNorm

    for n in range(1, nMax + 1):
        for m in range(n + 1):
            index = (n * (n + 1) // 2 + m);
            Pcup[index] = Pcup[index] * schmidtQuasiNorm[index];
            dPcup[index] = -dPcup[index] * schmidtQuasiNorm[index];
        # The sign is changed since the new WMM routines use derivative with respect to latitude
        # instead of co-latitude

    return [Pcup, dPcup]

def PcupHigh(lat: float, nMax: int):
    """This function evaluates all of the Schmidt-semi normalized associated Legendre
        functions up to degree nMax. The functions are initially scaled by
        10^280 sin^m in order to minimize the effects of underflow at large m
        near the poles (see Holmes and Featherstone 2002, J. Geodesy, 76, 279-299).
        Note that this function performs the same operation as MAG_PcupLow.
        However this function also can be used for high degree (large nMax) models.

        Calling Parameters:
                INPUT
                        nMax:	 Maximum spherical harmonic degree to compute.
                        x:		sin(latitude)

                OUTPUT
                        Pcup:	A vector of all associated Legendgre polynomials evaluated at
                                        x up to nMax. The lenght must by greater or equal to (nMax+1)*(nMax+2)/2.
                  dPcup:   Derivative of Pcup(x) with respect to latitude

                CALLS : none
        Notes:



      Adopted from the FORTRAN code written by Mark Wieczorek September 25, 2005.

      Manoj Nair, Nov, 2009 Manoj.C.Nair@Noaa.Gov

      Change from the previous version
      The prevous version computes the derivatives as
      dP(n,m)(x)/dx, where x = sin(latitude) (or cos(colatitude) ).
      However, the WMM Geomagnetic routines requires dP(n,m)(x)/dlatitude.
      Hence the derivatives are multiplied by sin(latitude).
      Removed the options for CS phase and normalizations.

      Note: In geomagnetism, the derivatives of ALF are usually found with
      respect to the colatitudes. Here the derivatives are found with respect
      to the latitude. The difference is a sign reversal for the derivative of
      the Associated Legendre Functions.

      The derivatives can't be computed for latitude = |90| degrees."""

    #double pm2, pm1, pmm, plm, rescalem, z, scalef;
    #double *f1, *f2, *PreSqr;
    #int k, kstart, m, n, NumTerms;
    x = math.sin(np.deg2rad(lat))
    NumTerms = int((nMax + 1) * (nMax + 2) / 2)

    z = math.sqrt((1.0 - x) * (1.0 + x))


    if z == 0:
        raise ValueError("z becomes error")

    f1 = np.zeros(NumTerms + 1)
    PreSqr = np.zeros(NumTerms + 1)
    f2 = np.zeros(NumTerms + 1)
    Pcup = np.zeros(NumTerms + 1)
    dPcup = np.zeros(NumTerms + 1)

    scalef = 1.0e-308

    for n in range(2 * nMax + 2):
        PreSqr[n] = math.sqrt(n * 1.0)

    k = 2
    for n in range(2, nMax + 1):
        k = k + 1
        f1[k] = float(2 * n - 1) / n
        f2[k] = float(n - 1) / n

        for m in range(1, n - 1):
            k = k + 1
            f1[k] = float(2 * n - 1) / PreSqr[n + m] / PreSqr[n - m]
            f2[k] = PreSqr[n - m - 1] * PreSqr[n + m - 1] / PreSqr[n + m] / PreSqr[n - m]

        k = k + 2

    pm2 = 1.0
    Pcup[0] = 1.0
    dPcup[0] = 0.0

    pm1 = x
    Pcup[1] = pm1
    dPcup[1] = z
    k = 1

    for n in range(2, nMax + 1):
        k = k + n
        plm = f1[k] * x * pm1 - f2[k] * pm2
        Pcup[k] = plm
        dPcup[k] = n * (pm1 - x * plm) / z
        pm2 = pm1
        pm1 = plm

    pmm = PreSqr[2] * scalef
    rescalem = 1.0 / scalef
    kstart = 0

    for m in range(1, nMax):
        rescalem = rescalem * z

        #Calculate Pcup(m, m)
        kstart = int(kstart + m + 1)
        pmm = pmm * PreSqr[2 * m + 1] / PreSqr[2 * m]
        Pcup[kstart] = pmm * rescalem / PreSqr[2 * m + 1]
        dPcup[kstart] = -float(m * x * Pcup[kstart] / z)
        pm2 = pmm / PreSqr[2 * m + 1]
        #Calculate Pcup(m + 1, m)
        k = kstart + m + 1
        pm1 = x * PreSqr[2 * m + 1] * pm2
        Pcup[k] = pm1 * rescalem
        dPcup[k] = float((pm2 * rescalem) * PreSqr[2 * m + 1] - x * (m + 1) * Pcup[k]) / z
        #Calculate Pcup(n, m)

        for n in range(m + 2, nMax + 1):
            k = k + n
            plm = x * f1[k] * pm1 - f2[k] * pm2
            Pcup[k] = plm * rescalem
            dPcup[k] = (PreSqr[n + m] * PreSqr[n - m] * (pm1 * rescalem) - float(n) * x * Pcup[k]) / z
            pm2 = pm1
            pm1 = plm

    #Calculate Pcup(nMax, nMax)

    rescalem = rescalem * z
    kstart = kstart + nMax + 1
    pmm = pmm / PreSqr[2 * nMax]
    Pcup[kstart] = pmm * rescalem
    dPcup[kstart] = -float(nMax * x * Pcup[kstart]) / z

    return Pcup, dPcup



import math
from typing import Optional, Tuple, Union

import numpy as np


def rad2deg(rad: float) -> float:
    """
        Convert radius to degree
    """

    return rad * 180.0 / math.pi


def deg2rad(deg) ->float:
    """
        Convert degree to radius
    """

    return deg * math.pi / 180.0


def calc_Bp_Pole(nmax: int, geoc_lat: Union[list[float], np.ndarray], sph:dict[str, list[float]], g: list[float], h: list[float]) -> float:
    """
    Calculate the B_phi magnetic elements at pole
    Args:
        nmax: maximum degree
        geoc_lat: geocentric latitude in degree
        sph: the dict svaed with spherical harmonic varialbles like (a/r) ^ (n+2), cos_m(lon), and sin_m(lon)
        g: g coefficients
        h: h coefficients

    Returns:

    """

    if isinstance(geoc_lat, list):
        geoc_lat = np.array(geoc_lat)
    PcupS = [0.0] * (nmax + 1)

    PcupS[0] = 1.0

    schmidtQuasiNorm1 = 1.0

    Bp = 0.0
    sin_phi = np.sin(deg2rad(geoc_lat))

    for n in range(1, nmax):
        idx = int(n * (n + 1) / 2 + 1)

        schmidtQuasiNorm2 = schmidtQuasiNorm1 * (2 * n - 1) / n
        schmidtQuasiNorm3 = schmidtQuasiNorm2 * np.sqrt((n * 2) / (n + 1))
        schmidtQuasiNorm1 = schmidtQuasiNorm2

        if n == 1:
            PcupS[1] = 1.0
        else:
            k = (((n - 1) * (n - 1)) - 1) / ((2 * n - 1) * (2 * n - 3))
            PcupS[n] = sin_phi * PcupS[n - 1] - k * PcupS[n - 2]

        Bp += sph["relative_radius_power"][n] * (g[idx] * sph["sin_mlon"][n] - h[idx] * sph["cos_mlon"][n]) * PcupS[
            n] * schmidtQuasiNorm3

    return Bp


def mag_SPH_summation(nmax: int, sph: dict[str, list[float]], g: list[float], h: list[float], Leg: list[list[float]],geoc_lat: Union[list[float], np.ndarray]) -> tuple:
    """
    Compute the magnetic eelements
    Args:
        nmax: max degree
        sph: the dict svaed with spherical harmonic varialbles like (a/r) ^ (n+2), cos_m(lon), and sin_m(lon)
        g: g coefficients
        h: h coefficients
        Leg: legendre function array. Leg[0] for Plm array; Leg[1] for dPlm array.
        geoc_lat: geocentric latitude in degree

    Returns:

    """

    if isinstance(geoc_lat, list):
        geoc_lat = np.array(geoc_lat)

    Br, Bt, Bp = np.zeros(len(geoc_lat)),np.zeros(len(geoc_lat)),np.zeros(len(geoc_lat))

    
    legP = np.array(Leg[0])
    legdP = np.array(Leg[1])
    
    pidx = 1

    for m in range(nmax + 1):
        # degree
        
        for n in range(m, nmax + 1):
            if n == 0:
                continue
            gidx = int(n * (n + 1) / 2 + m)
            Bt -= sph["relative_radius_power"][n] * (
                    g[gidx] * sph["cos_mlon"][m] + h[gidx] * sph["sin_mlon"][m]) * legdP[
                      pidx]

            Bp += sph["relative_radius_power"][n] * (
                    g[gidx] * sph["sin_mlon"][m] - h[gidx] * sph["cos_mlon"][m]) * m * legP[pidx]

            Br -= sph["relative_radius_power"][n] * (
                    g[gidx] * sph["cos_mlon"][m] + h[gidx] * sph["sin_mlon"][m]) * (
                          n + 1) * legP[pidx]
            pidx += 1

    cos_phi = np.cos(deg2rad(geoc_lat))

    mask = np.abs(cos_phi) < 1.0e-10 
    # Apply calc_Bp_Pole where the mask is True, otherwise perform division
    Bp = np.where(mask, Bp + calc_Bp_Pole(nmax, geoc_lat, sph, g, h), Bp / cos_phi)

    Bt = -Bt

    return Bt, Bp, Br


def mag_SPH_summation_alf(nmax, sph, coef_dict, legP, legdP, geoc_lat) -> tuple:
    """
    Compute the magnetic elements based on Legendre function from geomag's team C library
    Args:
        nmax:
        sph:
        coef_dict:
        legP:
        legdP:
        geoc_lat:

    Returns:

    """
    Br, Bt, Bp = 0.0, 0.0, 0.0

    for n in range(1, nmax + 1):
        # degree
        for m in range(n + 1):
            gidx = int(n * (n + 1) / 2 + m)

            Bt -= sph["relative_radius_power"][n] * (
                    coef_dict["g"][gidx] * sph["cos_mlon"][m] + coef_dict["h"][gidx] * sph["sin_mlon"][m]) * legdP[
                      gidx]

            Bp += sph["relative_radius_power"][n] * (
                    coef_dict["g"][gidx] * sph["sin_mlon"][m] - coef_dict["h"][gidx] * sph["cos_mlon"][m]) * m * legP[
                      gidx]

            Br -= sph["relative_radius_power"][n] * (
                    coef_dict["g"][gidx] * sph["cos_mlon"][m] + coef_dict["h"][gidx] * sph["sin_mlon"][m]) * (
                          n + 1) * legP[gidx]

    cos_phi = np.cos(deg2rad(geoc_lat))

    mask = np.abs(cos_phi) < 1.0e-10 
    # Apply calc_Bp_Pole where the mask is True, otherwise perform division
    Bp = np.where(mask, Bp + calc_Bp_Pole(nmax, geoc_lat, sph, g, h), Bp / cos_phi)


    return Bt, Bp, Br


def rotate_magvec(Bt, Bp, Br, geoc_lat, geod_lat) -> Tuple[float, float, float]:
    """
            Convert magnetic vector from spherical to geodetic

            Parameters:
            ___________

            Bt: magnetic elements theta
            Bp: magnetic elements phi
            Br: magnetic elements radius
            geoc_lat: geocentric latitude
            geod_lat: geeodetic latitude

            Returns:
            _________

            B:array the magnetic vector based on geodetic
            B = [Bx, By, Bz]
    """

    psi = (math.pi / 180.0) * (geoc_lat - geod_lat)

    Bz = Bt * np.sin(psi) + Br * np.cos(psi)
    Bx = Bt * np.cos(psi) - Br * np.sin(psi)
    By = Bp

    return Bx, By, Bz


class GeomagElements:
    def __init__(self, 
                 Bx: Union[float, np.ndarray, list],
                 By: Union[float, np.ndarray, list],
                 Bz: Union[float, np.ndarray, list],
                 dBx: Optional[Union[float, np.ndarray, list]] = None,
                 dBy: Optional[Union[float, np.ndarray, list]] = None,
                 dBz: Optional[Union[float, np.ndarray, list]] = None):
        """
        Compute magnetic elements.
        
        Args:
            Bx: float or np.ndarray
            By: float or np.ndarray
            Bz: float or np.ndarray
            dBx: Optional float or np.ndarray
            dBy: Optional float or np.ndarray
            dBz: Optional float or np.ndarray
        """
        self.Bx = np.asarray(Bx, dtype=np.float64)
        self.By = np.asarray(By, dtype=np.float64)
        self.Bz = np.asarray(Bz, dtype=np.float64)

        self.dBx = np.asarray(dBx, dtype=np.float64) if dBx is not None else None
        self.dBy = np.asarray(dBy, dtype=np.float64) if dBy is not None else None
        self.dBz = np.asarray(dBz, dtype=np.float64) if dBz is not None else None



    def get_Bh(self) -> float:
        """
            Compute the magnetic horizontal


            Returns:
            ____________

            h: magneitc horizontal elements

        """

        return np.sqrt(self.Bx ** 2 + self.By ** 2)

    def get_Bf(self) -> float:
        """
            Get the total intensity
            Returns:
            f: the total intensity value
            _________
        """

        f = np.sqrt(self.Bx ** 2 + self.By ** 2 + self.Bz ** 2)
        return f

    def get_Bdec(self) -> float:
        """
        Get the declination value
        """
        dec = rad2deg(np.arctan2(self.By, self.Bx))

        return dec

    def get_Binc(self) -> float:
        """
        Get the inclination value
        Returns:

        """
        Bh = self.get_Bh()
        inc = rad2deg(np.arctan2(self.Bz, Bh))

        return inc

    def get_all_base(self) -> dict[str, float]:
        """
        Get Bx, By, Bz, Bh, Bf, Bdec and Binc in dict

        """
    
        mag_map = {}

        mag_map["x"] = np.asarray(self.Bx, dtype=np.float64)
        mag_map["y"] = np.asarray(self.By, dtype=np.float64)
        mag_map["z"] = np.asarray(self.Bz, dtype=np.float64)
        mag_map["h"] = np.asarray(self.get_Bh(), dtype=np.float64)
        mag_map["f"] = np.asarray(self.get_Bf(), dtype=np.float64)
        mag_map["dec"] = np.asarray(self.get_Bdec(), dtype=np.float64)
        mag_map["inc"] = np.asarray(self.get_Binc(), dtype=np.float64)
        return mag_map

    def get_all(self) -> dict[str, float]:
        """

        Returns: all of magnetic elements:
        Bx, By, Bz, Bh, Bf, Bdec, Binc,
        dBx, dBy, dBz, dBh, dBf, dBdec and dBinc in dict

        """
        mag_map = {}

        mag_map["x"] = np.asarray(self.Bx, dtype=np.float64)
        mag_map["y"] = np.asarray(self.By, dtype=np.float64)
        mag_map["z"] = np.asarray(self.Bz, dtype=np.float64)
        h = np.asarray(self.get_Bh(), dtype=np.float64)
        f = np.asarray(self.get_Bf(), dtype=np.float64)

        mag_map["h"] = h
        mag_map["f"] = f
        mag_map["dec"] = self.get_Bdec()
        mag_map["inc"] = self.get_Binc()

        mag_map["dx"] = self.dBx
        mag_map["dy"] = self.dBy
        mag_map["dz"] = self.dBz
        mag_map["dh"] = (self.Bx * self.dBx + self.By * self.dBy) / h
        mag_map["df"] = (self.Bx * self.dBx + self.By * self.dBy + mag_map["z"] * self.dBz) / mag_map["f"]
        mag_map["ddec"] = 180 / math.pi * (self.Bx * self.dBy - self.By * self.dBx) / (h ** 2) / 60
        mag_map["dinc"] = np.asarray((180 / math.pi * (h * self.dBz - self.Bz * mag_map["dh"])) / (f ** 2), dtype=np.float64)/ 60
        
        return mag_map

    def get_dBh(self) -> float:
        """

        Returns: delta horizontal

        """
        h = self.get_Bh()
        return (self.Bx * self.dBx + self.By * self.dBy) / h

    def get_dBf(self) -> float:
        """
        Returns: delta total intensity
        """
        f = self.get_Bf()
        return (self.Bx * self.dBx + self.By * self.dBy + self.Bz * self.dBz) / f

    def get_dBdec(self) -> float:
        """
        Returns: delta declination value
        """
        h = self.get_Bh()
        return 180 / math.pi * (self.Bx * self.dBy - self.By * self.dBx) / (h ** 2)

    def get_dBinc(self) -> float:
        """
        Returns: delta inclination value

        """
        h = self.get_Bh()
        f = self.get_Bf()
        dh = (self.Bx * self.dBx + self.By * self.dBy) / h
        return 180 / math.pi * (h * self.dBz - self.Bz * dh) / (f ** 2)
